package com.nokia.nfc.app.test.manual;

import javax.microedition.lcdui.Display;
import javax.microedition.midlet.MIDlet;

import com.nokia.nfc.app.test.manual.ui.*;
import com.nokia.nfc.app.util.Log;
import com.nokia.nfc.app.util.Settings;

public class Control {
    
    private static final String SET_FILE_LOGGING = "FileLogging";
    
    private static final String SET_CONSOLE_LOGGING = "ConsoleLogging";
    
    private static final long SPLASH_MIN_TIME_VISIBLE = 1000;

    private MIDlet midlet;
    
    private ViewHandler viewHandler;

    
    /**
     * Constructs the Control object.
     */
    public Control(MIDlet midlet) {
        this.midlet = midlet;
    }
    
    /**
     * Starts the initialization thread. Splash screen must be set visible 
     * in the MIDlet's startApp() method before calling this method.
     */
    public void start() {
        new Thread() {
            public void run() {
                initialize();
            }
        }.start();
    }
    
    private void initialize() {
        long t0 = System.currentTimeMillis();
        
        // Put all initialization stuff here.
        Settings.initialize(midlet);
        
        // Read default log settings from Jad attributes.
        if (isEnabled("FileLogging", false)) {
            setFileLogging(true, false);
        } else {
            setFileLogging(false, false);
        }
        
        if (isEnabled("ConsoleLogging", false)) {
            setConsoleLogging(true, false);
        } else {
            setConsoleLogging(false, false);
        }

        Display display = Display.getDisplay(midlet);
        DisplayColors.initialize(display);
        ViewHandler.initialize(display, this);
        viewHandler = ViewHandler.getInstance();
        
        long t1 = System.currentTimeMillis();
        long remaining = SPLASH_MIN_TIME_VISIBLE - (t1 - t0);
        
        // Ensure that the splash is shown for at least the specified minimum time.
        if (remaining > 0) {
            try {
                Thread.sleep(SPLASH_MIN_TIME_VISIBLE);
            } catch (InterruptedException e) {
            }
        }

        viewHandler.display(new TestSelector(this, NFCTest.TESTS));
    }
    
    /**
     * Checks the value of a Jad attribute and returns the corresponding 
     * boolean value.
     * Only attributes that have <code>enabled</code> and <code>disabled</code>
     * should be used with this method.
     * 
     * @param attrName name of the attribute to check.
     * @param defaultValue default value to be used in case the attribute is not defined.
     * 
     * @return <code>true</code> if the specified attribute has the value 
     *         <code>enabled</code> and <code>false</code> otherwise.
     */
    public boolean isEnabled(String attrName, boolean defaultValue) {
        String value = Settings.getSetting(attrName, true);
        
        if (value == null) {
            return defaultValue;
        } else if (value.equalsIgnoreCase("enabled")) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * Releases all resources. Only {@link NFCTest} should call
     * this when exiting the MIDlet due to user pressing the end call key. 
     */
    void cleanUp() {
        ViewHandler.cleanUp();
        
        try {
            Log.setFileLogger(false, null);
        } catch (Exception e) {
        } 
    }
    
    /**
     * Causes the application to exit and also calls {@link #cleanUp()}.
     */
    public void exit() {
        cleanUp();
        midlet.notifyDestroyed();
    }
    
    /**
     * Enable or disable file logging.
     * 
     * @param enabled if <code>true</code> then file logging will be enabled.
     * @return enabled/disabled status after trying to change status.
     */
    public boolean setFileLogging(boolean enabled, boolean save) {

        try {
            Log.setFileLogger(enabled, "NFCTest"); 
            
            if (save) {
                saveLoggingSetting(SET_FILE_LOGGING, enabled);
            }
            
        } catch (Exception e) {
            // Failed to open file log (probably because the user gave no permissions)
        }

        return Log.isUsingFiles();
    }
    
    private static final void saveLoggingSetting(String name, boolean enabled) {
        Settings.setSetting(name, enabled ? "enabled" : "disabled");
    }
    
    /**
     * Enable or disable console logging.
     * 
     * @param enabled if <code>true</code> then console logging will be enabled.
     */
    public void setConsoleLogging(boolean enabled, boolean save) {
        Log.setConsoleLogger(enabled);
        
        if (save) {
            saveLoggingSetting(SET_CONSOLE_LOGGING, enabled);
        }
    }
    
}
