package com.nokia.nfc.app.test.manual.tests;

import java.io.IOException;
import java.util.Enumeration;
import java.util.Hashtable;

import javax.microedition.lcdui.*;

import com.nokia.nfc.app.test.manual.ui.TestForm;
import com.nokia.nfc.app.test.manual.ui.ViewHandler;
import com.nokia.nfc.app.util.connector.ConnectionFactory;
import com.nokia.nfc.llcp.*;

/**
 * Enables sending and receiving data using {@link LLCPConnection}.
 */
public class LLCP extends TestForm implements LLCPLinkListener, 
    ErrorRecoveryListener, LLCPConnectionListener 
{
    
    private TextField uidField;
    private TextField errRecoveryStatusField;
    private ChoiceGroup connSelector;
    private TextField sendField;
    private StringItem recvField;
    private LLCPManager llcpMngr;
    private Hashtable connections;
    
    // Up to 100 connections can be listened to simultaneously. 
    // Arrays contain Type&PID pairs.
    private byte[][] listeningToConnections = new byte[100][];
    
    
    public LLCP() {
        super("LLCP");
    }

    /**
     * @see com.nokia.nfc.app.test.manual.ui.TestForm#open()
     */
    public void open() {
        super.open();
        
        try {
            llcpMngr = LLCPManager.getInstance();
        } catch (Exception e) {
            log("Failed to get LLCPManager instance", e);
            return;
        }
        
        connections = new Hashtable();
        addOkCmd("Open connection");
        
        uidField = new TextField("Remote device UID", "link down", 20, TextField.ANY | TextField.UNEDITABLE);
        errRecoveryStatusField = new TextField("Error recovery status", null, 20, TextField.ANY | TextField.UNEDITABLE);

        connSelector = new ChoiceGroup("Select connection", Choice.POPUP);
        sendField = new TextField("Data to send", "", 70000, TextField.ANY);
        recvField = new StringItem("Received data", "", StringItem.PLAIN); 
        
        append(uidField);
        append(errRecoveryStatusField);
        append(connSelector);
        append(sendField);
        append(recvField);
        addOkButton("Send");
        addOkButton("Close");
        
        llcpMngr.addErrorRecoveryListener(this);
        llcpMngr.addLinkListener(this);
    }
    
    /**
     * Close all open LLCP connections.
     * 
     * @see com.nokia.nfc.app.test.manual.ui.TestForm#close()
     */
    public void close() {
        super.close();
        
        if (connections != null) {
            Enumeration connEnum = connections.elements();
            
            while (connEnum.hasMoreElements()) {
                LLCPConnection conn = (LLCPConnection) connEnum.nextElement();
                
                try {
                    conn.close();
                } catch (IOException e) {
                }
            }
            
            connections.clear();
        }
        
        if (llcpMngr != null) {
            llcpMngr.removeErrorRecoveryListener(this);
            llcpMngr.removeLinkListener(this);
        }
        
        stopListeningToConnections();
    }
    
    /**
     * @return the selected connection or <code>null</code> if no connection
     *         is currently selected.
     */
    private LLCPConnection getSelectedConnection() {
        LLCPConnection conn = null;
        int selectedIndex = connSelector.getSelectedIndex();
        
        if (selectedIndex > -1) {
            String key = connSelector.getString(selectedIndex);
            conn = (LLCPConnection) connections.get(key);
        } else {
            log("Select a connection first");
        }
        
        return conn;
    }
    
    /**
     * @see com.nokia.nfc.app.test.manual.ui.TestForm#handleOkCmd(java.lang.String)
     */
    protected void handleOkCmd(String label) {
        
        if ("Open connection".equals(label)) {
            showOpenConnectionView();
        }
        else if ("Send".equals(label)) {

            try {
                String str = sendField.getString();
                byte[] data = str.getBytes();
                log("SEND (" + data.length + "B): " + str);
                LLCPConnection conn = getSelectedConnection();
                conn.send(data);
            } catch (IOException e) {
                log("Send failed", e);
            }
        }
        else if ("Close".equals(label)) {
            // Assumes that the selection doesn't change during this operation..
            LLCPConnection conn = getSelectedConnection();
            
            if (conn != null) {
                try {
                    log("Close connection (" + conn.toString() + ")");
                    conn.close();
                    connSelector.delete(connSelector.getSelectedIndex());
                } catch (IOException e) {
                    log("Failed to close connection", e);
                }
            }
        }
    }
    
    /**
     * Display a view where the user can define the transport type and
     * protocol identifier and then either open a connection or start
     * listening for one.
     */
    private void showOpenConnectionView() {
        OpenConnectionView ocv = new OpenConnectionView();
        ViewHandler.getInstance().display(ocv);
    }
    
    /**
     * Called to stop listening to connections that this object is currently
     * listening to.
     */
    private void stopListeningToConnections() {

        for (int i = 0; i < listeningToConnections.length; i++) {
            byte[] listenInfo = listeningToConnections[i];
            
            if (listenInfo != null) {
                byte transportType = listenInfo[0];
                byte pid = listenInfo[1];
                llcpMngr.stopListening(transportType, pid, this);
                listeningToConnections[i] = null;
            }
        }
    }
    
    /**
     * Called to start listening for a {@link LLCPConnection}.
     */
    private void listenToConnection(byte transportType, byte pid) {
        boolean foundSlot = false;
        
        for (int i = 0; i < listeningToConnections.length; i++) {
            
            if (listeningToConnections[i] == null) {
                listeningToConnections[i] = new byte[] {transportType, pid};
                foundSlot = true;
                break;
            }
        }
        
        if (foundSlot) {
            llcpMngr.startListening(transportType, pid, this);
            log("Started listening to type" + transportType + ", pid=" + pid);
        } else {
            log("No more connection listeners allowed");
        }
    }

    /**
     * Called when a {@link LLCPConnection} has been opened or if a connection
     * should be opened (<code>conn</code> is <code>null</code>).
     */
    private void openConnection(byte transportType, byte pid, LLCPConnection conn) {
        
        if (conn == null) {
            String uri = "nfc:llcp;type=" + transportType + ";pid=" + pid + ";uid=" + uidField.getString();
            log("Open connection", uri);
            
            try {
                conn = (LLCPConnection) ConnectionFactory.getInstance().open(uri);
            } catch (IOException e) {
                log("Failed to open connection", e);
            }
        }
        
        if (conn != null) {
            String baseKey = "type" + transportType + ", pid=" + pid;
            String key = baseKey;
            int i = 1;
            
            while (connections.containsKey(key)) {
                key = baseKey + " (" + i++ + ")";
            }
            
            connections.put(key, conn);
            new Thread(new ReceiverThread(conn)).start();
            connSelector.append(key, null);
        }
    }

    /**
     * @see com.nokia.nfc.llcp.LLCPLinkListener#linkEstablished(java.lang.String)
     */
    public void linkEstablished(String uid) {
        uidField.setString(uid);
    }
    
    /**
     * @see com.nokia.nfc.llcp.LLCPLinkListener#linkLost()
     */
    public void linkLost() {
        uidField.setString("link down");
        errRecoveryStatusField.setString(null);
    }

    /**
     * @see com.nokia.nfc.llcp.ErrorRecoveryListener#recoveryStatus(int)
     */
    public void recoveryStatus(int status) {

        switch (status) {
            case ErrorRecoveryListener.RECOVERY_STARTED:
                errRecoveryStatusField.setString("Started");
                break;
                
            case ErrorRecoveryListener.RECOVERY_FAILED:
                errRecoveryStatusField.setString("Failed");
                break;
                
            case ErrorRecoveryListener.RECOVERY_SUCCESSFUL:
                errRecoveryStatusField.setString("Successful");
                break;
                
            default:
                errRecoveryStatusField.setString("Unknown: " + status);
                break;
        }
    }
    
    /**
     * @see com.nokia.nfc.llcp.LLCPConnectionListener#connectionOpened(com.nokia.nfc.llcp.LLCPConnection)
     */
    public void connectionOpened(LLCPConnection connection) {
        log("connectionOpened called (" + connection.toString() + ")");

        openConnection(connection.getTransportType(), connection.getProtocolId(), connection);
    }
    
    /**
     * This thread is used for receiving data in the background until the 
     * connection is closed or some error occurrs.
     */
    class ReceiverThread implements Runnable {
        
        private LLCPConnection conn;
        
        private ReceiverThread(LLCPConnection conn) {
            this.conn = conn;
        }
        
        /**
         * @see java.lang.Runnable#run()
         */
        public void run() {
            
            while (conn != null) {
                try {
                    byte[] receivedData = conn.receive();
                    String str = new String(receivedData);
                    
                    if (getSelectedConnection() == conn) {
                        recvField.setText(str);
                    }
                    
                    log("RECV (" + receivedData.length + "B): " + str);                    
                } catch (IOException e) {
                    log("Receive failed", e);
                    break;
                }
            }
        }
    }

    class OpenConnectionView extends TestForm {
        
        private ChoiceGroup typeSelector;
        private TextField pidField;
        
        private OpenConnectionView() {
            super("Open connection");
            
            typeSelector = new ChoiceGroup("Transport type", Choice.POPUP);
            typeSelector.append("Type 1", null);
            typeSelector.append("Type 2", null);
            pidField = new TextField("Protocol ID", "", 4, TextField.NUMERIC);
            
            append(typeSelector);
            append(pidField);
            
            addOkButton("Connect");
            addOkButton("Listen");
        }
        
        /**
         * @see com.nokia.nfc.app.test.manual.ui.TestForm#open()
         */
        public void open() {
            super.open();
        }
        
        /**
         * @see com.nokia.nfc.app.test.manual.ui.TestForm#close()
         */
        public void close() {
            super.close();
        }
        
        /**
         * @see com.nokia.nfc.app.test.manual.ui.TestForm#handleOkCmd(java.lang.String)
         */
        protected void handleOkCmd(String label) {
            byte type = 0;
            byte pid = 0;

            try {
                type = (byte) (typeSelector.getSelectedIndex() + 1);
                pid = Byte.parseByte(pidField.getString());
            } catch (Exception e) {
                log("Failed to read connection parameters", e);
                return;
            }
            
            if ("Listen".equals(label)) {
                listenToConnection(type, pid);
            }
            else if ("Connect".equals(label)) {
                openConnection(type, pid, null);
            }
            
            ViewHandler.getInstance().displayPrevious();
        }
    }

}
