package com.nokia.nfc.app.test.manual.ui;

import java.io.IOException;
import java.util.Vector;

import javax.microedition.lcdui.*;
import javax.microedition.lcdui.game.Sprite;

import com.nokia.nfc.app.util.*;

/**
 * Item used for displaying N last lines of log output. The amount N is defined
 * in the constructor. The displayed amount can be smaller if wrapping is enabled.
 * The item height is constant.
 */
public class LogItem extends CustomItem implements LogListener, ItemCommandListener {
    
    /** Minimum allowed tail length. */
    public static final int MIN_TAIL_LENGTH = 1;
    
    /** Maximum allowed tail length. */
    public static final int MAX_TAIL_LENGTH = 10;
    
    /** Images used for drawing the background. */
    private static Image hilightedCorner;
    private static Image hilightedBlock;
    private static Image block;
    private static Image corner;
    
    /** Top, right, bottom and left margin. */
    private static final int MARGIN = 5;
    
    /** Command for opening the log in fullscreen mode. */
    private Command openLogCmd = new Command("Open", Command.OK, 1);
    
    /** Number of latest log lines to display. */
    private int tailLength;
    
    /** Height of the item. */
    private int height;
    
    /** Width of the item. */
    private final int width;
    
    /** Font used for drawing log lines. */
    private Font font;
    
    /** Defines whether this item is selected (highlighted) or not. */
    private boolean selected;
    
    /** If <code>true</code> then line wrapping is enabled. */
    private boolean wrapping;
    
    
    /**
     * @param tailLength (maximum) amount of last log lines to display.
     * @param width width reserved for this item.
     */
    public LogItem(int tailLength, int width) {
        super("Log");
        this.tailLength = tailLength;
        
        if (tailLength < MIN_TAIL_LENGTH || tailLength > MAX_TAIL_LENGTH) {
            throw new IllegalArgumentException("Invalid tailLength");
        }
        
        this.wrapping = true;
        this.font = Font.getFont(Font.FACE_PROPORTIONAL, Font.STYLE_PLAIN, Font.SIZE_SMALL);
        this.height = font.getHeight() * tailLength + MARGIN * 2;
        this.width = width;
        this.selected = false;
        
        setItemCommandListener(this);
        addCommand(openLogCmd);
    }
    
    /**
     * @see javax.microedition.lcdui.ItemCommandListener#commandAction(javax.microedition.lcdui.Command, javax.microedition.lcdui.Item)
     */
    public void commandAction(Command c, Item item) {
        
        if (c == openLogCmd) {
            ViewHandler.getInstance().display(new LogView());
        }
    }
    
    /**
     * @see com.nokia.nfc.app.util.LogListener#lineAdded(java.lang.String)
     */
    public void lineAdded(String line) {
        repaint();
    }
    
    /**
     * @see javax.microedition.lcdui.CustomItem#traverse(int, int, int, int[])
     */
    protected boolean traverse(int dir, int viewportWidth, int viewportHeight, int[] visRect_inout) {

        if (selected) {
            return false;
        } else {
            selected = true;
            return true;
        }
    }
    
    /**
     * @see javax.microedition.lcdui.CustomItem#traverseOut()
     */
    protected void traverseOut() {
        selected = false;
    }

    /**
     * @see javax.microedition.lcdui.CustomItem#getMinContentHeight()
     */
    protected int getMinContentHeight() {
        return height;
    }

    /**
     * @see javax.microedition.lcdui.CustomItem#getMinContentWidth()
     */
    protected int getMinContentWidth() {
        return width;
    }

    /**
     * @see javax.microedition.lcdui.CustomItem#getPrefContentHeight(int)
     */
    protected int getPrefContentHeight(int width) {
        return height;
    }

    /**
     * @see javax.microedition.lcdui.CustomItem#getPrefContentWidth(int)
     */
    protected int getPrefContentWidth(int height) {
        return width;
    }

    /**
     * @see javax.microedition.lcdui.CustomItem#paint(javax.microedition.lcdui.Graphics, int, int)
     */
    protected void paint(Graphics g, int w, int h) {
        drawBackground(g, w, h);
        
        String[] lines = Log.getBackBuffer().tail(tailLength);
        g.setFont(font);
        g.setColor(DisplayColors.getColor(selected ? Display.COLOR_HIGHLIGHTED_FOREGROUND : Display.COLOR_FOREGROUND));
        
        int y = h - MARGIN;
        int x = MARGIN;
        final int textW = width - 2 * MARGIN;
        final int textH = height - 2 * MARGIN;
        int anchor = Graphics.BOTTOM | Graphics.LEFT;
        g.setClip(x, MARGIN, textW, textH);
        
        if (wrapping) {
            lines = wrapLines(lines, textW, font);
        } else {
            lines = removeNulls(lines);
        }
        
        // wrapLines / removeNulls both remove null lines.
        
        if (lines.length < tailLength) {
            y -= (tailLength - lines.length) * font.getHeight();
        }
        
        for (int i = lines.length - 1; i >= 0 && y > MARGIN; i--) {
            g.drawString(lines[i], x, y, anchor);
            y -= font.getHeight();
        }
    }
    
    private final static String[] removeNulls(String[] lines) {
        int first = -1;
        int last = -1;
        
        for (int i = 0; i < lines.length; i++) {

            if (lines[i] != null) {
                if (first == -1) {
                    first = i;
                } 
                
                if (last == -1) {
                    last = i;
                }
            }
        }
        
        int len = last - first + 1;
        String[] ret = new String[len];
        System.arraycopy(lines, first, ret, 0, len);
        return ret;
    }
    
    private final static String[] wrapLines(String[] lines, int w, Font font) {
        Vector newLines = new Vector();
        int offset;
        int remaining;
        char[] chars;
        
        for (int i = 0; i < lines.length; i++) {
            offset = 0;
            String line = lines[i];
            
            if (line == null) {
                continue;
            }
            
            if (font.stringWidth(line) > w) {
                // Truncate line into pieces.
                chars = line.toCharArray();
                remaining = chars.length;
                
                while (remaining > 0) {
                    int len = 1;
                    
                    for (; (offset + len) < chars.length; len++) {
                        
                        if (font.charsWidth(chars, offset, len) > w) {
                            len--;
                            break;
                        }
                    }
                    
                    newLines.addElement(line.substring(offset, offset + len));
                    offset += len;
                    remaining -= len;
                }
                
            } else {
                newLines.addElement(line);
            }
        }
        
        String[] ret = new String[newLines.size()];
        newLines.copyInto(ret);
        return ret;
    }
    
    /**
     * Draws the item background using small images.
     */
    private final void drawBackground(Graphics g, int w, int h) {
        initializeImages();
        
        Image corner;
        Image block;
        
        if (selected) {
            corner = hilightedCorner;
            block = hilightedBlock;
        } else {
            corner = LogItem.corner;
            block = LogItem.block;
        }
         
        int cWidth = corner.getWidth();
        int cHeight = corner.getHeight();
        int bWidth = block.getWidth();
        int bHeight = block.getHeight();
        // It is assumed that bHeight == cHeight. Otherwise bottom part of 
        // the top corners would not get drawn.
        
        // Draw all corners first
        g.drawRegion(corner, 0, 0, cWidth, cHeight, Sprite.TRANS_NONE, 0, 0, Graphics.TOP | Graphics.LEFT);
        g.drawRegion(corner, 0, 0, cWidth, cHeight, Sprite.TRANS_ROT90, w - 1, 0, Graphics.TOP | Graphics.RIGHT);
        g.drawRegion(corner, 0, 0, cWidth, cHeight, Sprite.TRANS_ROT180, w - 1, h - 1, Graphics.BOTTOM | Graphics.RIGHT);
        g.drawRegion(corner, 0, 0, cWidth, cHeight, Sprite.TRANS_ROT270, 0, h - 1, Graphics.BOTTOM | Graphics.LEFT);
        
        // Then fill the remaining area with blocks.
        int x = cWidth;
        int y = 0;
        int drawWidth;
        int drawHeight = bHeight;
        int remainingHeight = h;

        while (remainingHeight > 0) {
            boolean cornerfulRow = false;
            int remainingWidth = w;

            if (y == 0 || y == h - cHeight) {
                // Either first or last row. Don't draw on top of the corners.
                x = cWidth;
                remainingWidth -= cWidth * 2;
                cornerfulRow = true;
            } else {
                x = 0;
            }
            
            drawWidth = bWidth;
            
            while (remainingWidth > 0) {
                g.drawRegion(block, 0, 0, drawWidth, drawHeight, Sprite.TRANS_NONE, x, y, Graphics.TOP | Graphics.LEFT);
                remainingWidth -= drawWidth;
                x += drawWidth;
                
                if (cornerfulRow && remainingWidth < drawWidth) {
                    // Don't draw on top of the corner's left hand side.
                    drawWidth = remainingWidth;
                }
            }
            
            remainingHeight -= drawHeight;
            y += drawHeight;
            
            if (remainingHeight < (drawHeight + cHeight)) {
                // Second last row. Don't draw on top of the corner's upper part.
                drawHeight = remainingHeight - cHeight;
            } else if (drawHeight < bHeight) {
                // Last row. Resume correct height.
                drawHeight = bHeight;
            } 
        }
    }
    
    /**
     * Loads all the resource image files used in drawing.
     */
    private synchronized final void initializeImages() {
        
        if (hilightedBlock == null) {
            
            try {
                Image origBlock = Image.createImage("/block.png");
                Image origCorner = Image.createImage("/corner.png");
                int[] origBlockRGB = new int[origBlock.getWidth() * origBlock.getHeight()];
                int[] origCornerRGB = new int[origCorner.getWidth() * origCorner.getHeight()];
                origBlock.getRGB(origBlockRGB, 0, origBlock.getWidth(), 0, 0, origBlock.getWidth(), origBlock.getHeight());
                origCorner.getRGB(origCornerRGB, 0, origCorner.getWidth(), 0, 0, origCorner.getWidth(), origCorner.getHeight());
                
                block = colorRGB(origBlockRGB, DisplayColors.getColor(Display.COLOR_BACKGROUND), 
                        origBlock.getWidth(), origBlock.getHeight());
                corner = colorRGB(origCornerRGB, DisplayColors.getColor(Display.COLOR_BACKGROUND), 
                        origCorner.getWidth(), origCorner.getHeight());
                hilightedBlock = colorRGB(origBlockRGB, DisplayColors.getColor(Display.COLOR_HIGHLIGHTED_BACKGROUND), 
                        origBlock.getWidth(), origBlock.getHeight());
                hilightedCorner = colorRGB(origCornerRGB, DisplayColors.getColor(Display.COLOR_HIGHLIGHTED_BACKGROUND), 
                        origCorner.getWidth(), origCorner.getHeight());
            } catch (IOException ioe) {
                if (LogLevel.level >= Log.ERROR) Log.log("Failed to load LogItem images", ioe);
            }
        }
    }
    
    /**
     * Modifies the given ARGB data by replacing RGB value with the parameter
     * <code>color</code>.
     */
    private static final Image colorRGB(int[] src, int color, int width, int height) {
        int alpha;
        
        for (int i = 0; i < src.length; i++) {
            alpha = src[i] & 0xFF000000;
            src[i] = alpha | (color & 0x00FFFFFF);
        }
        
        return Image.createRGBImage(src, width, height, true);
    }
    
}
