package com.nokia.nfc.app.test.manual.ui;

import javax.microedition.lcdui.*;

import com.nokia.nfc.app.util.*;

/**
 * <p>TextField class extension that provides two new additional features. First
 * one can select from a list of previous values. Method {@link #storeValue()} must
 * be called to store the current string for later selection. Second feature
 * enables reading input from COMM port. When this is enabled and the input
 * is received from COMM port then received input is appended to the end
 * of the current string. The user must explicitly turn on the COMM port
 * connection from the UI.</p> 
 */
public class SelectableTextField extends TextField implements ItemCommandListener, InputListener {

    private Command chooseCmd;
    
    private Command openCommCmd;
    
    private Command closeCommCmd;
    
    private Command clearCmd;
    
    private String[] previousValues;
    
    private boolean selectable;
    
    private CommPortInput commInput;
    
    private boolean ignoreCase = false;
    
    private Font selectionFont = null;
    
    private String id;
    
    // TODO add possibility to enable|disable autosaving by id & saving the setting permanently.
    
    
    /**
     * Both previous value selection and COMM port input is enabled. Constraints
     * define that predictive input is disabled.
     */
    public SelectableTextField(String label, String text, int maxSize) {
        this(label, text, maxSize, TextField.ANY | TextField.NON_PREDICTIVE, true, true);
    }
    
    /**
     * @param selectable if <code>true</code> then the user can select from 
     *                   a list of previous values (if any).
     * @param enableCommInput if <code>true</code> then the user can choose to 
     *                       enable reading input from a COMM port. 
     */
    public SelectableTextField(String label, String text, int maxSize, int constraints, 
            boolean selectable, boolean enableCommInput)
    {
        this(label, text, maxSize, constraints, selectable, enableCommInput, null, null);
    }
    
    /**
     * @param selectable if <code>true</code> then the user can select from 
     *                   a list of previous values (if any).
     * @param enableCommInput if <code>true</code> then the user can choose to 
     *                       enable reading input from a COMM port.
     * @param initialValues used as the selectable values in case there.  
     * @param id the identifier of this field - use this in case the label cannot be used as a valid filename.
     */
    public SelectableTextField(String label, String text, int maxSize, int constraints, 
            boolean selectable, boolean enableCommInput, String[] initialValues, String id) 
    {
        super(label, text, maxSize, constraints);
        this.selectable = selectable;
        
        if (id == null) {
            id = label;
        }
        
        this.id = id;
        
        if (selectable) {
            chooseCmd = new Command("Choose", Command.ITEM, 2);
            String[] values = StringValuesManager.loadGroupValues(getId(), initialValues);
            setPreviousValues(values);
        }

        clearCmd = new Command("Clear", Command.ITEM, 10);
        addCommand(clearCmd);

        if (enableCommInput) {
            openCommCmd = new Command("Enable COMM input", Command.ITEM, 20);
            closeCommCmd = new Command("Disable COMM input", Command.ITEM, 1);
            addCommand(openCommCmd);
        }
        
        setItemCommandListener(this);
    }
    
    public String getId() {
        return this.id;
    }
    
    /**
     * Set the font to be used in the previous values selection list.
     */
    public void setSelectionFont(Font font) {
        this.selectionFont = font;
    }
    
    /**
     * If parameter value is <code>true</code> then the case of the string's
     * characters will be ignored when storing the value.
     */
    public void setIgnoreCase(boolean enabled) {
        ignoreCase = enabled;
    }
    
    public void cleanUp() {
        
        if (commInput != null) {
            commInput.close();
        }
    }
    
    /**
     * @see com.nokia.nfc.app.util.InputListener#notifyInput(java.lang.String, java.lang.Object)
     */
    public synchronized void notifyInput(String input, Object src) {
        String current = getString();
        int curLen = current.length();
        int addLen = input.length();
        String combined = null;
        
        if ((curLen + addLen) > getMaxSize()) {
            combined = current + input.substring(0, getMaxSize() - curLen);
        } else {
            combined = current + input;
        }
        
        setString(combined);   
    }
    
    private void setPreviousValues(String[] previousValues) {
        this.previousValues = previousValues;
        
        if (previousValues != null && previousValues.length > 0) {
            addCommand(chooseCmd);
        } else {
            removeCommand(chooseCmd);
        }
    }

    /**
     * @see javax.microedition.lcdui.ItemCommandListener#commandAction(javax.microedition.lcdui.Command, javax.microedition.lcdui.Item)
     */
    public void commandAction(Command c, Item item) {
        
        if (c == chooseCmd) {
            PreviousValueSelection pvs = new PreviousValueSelection(previousValues);
            ViewHandler.getInstance().display(pvs);
        } else if (c == openCommCmd) {
            commInput = CommPortInput.createDefaultInput();
            
            if (commInput.open()) {
                commInput.addListener(this);
                removeCommand(openCommCmd);
                addCommand(closeCommCmd);
            } 
            
        } else if (c == closeCommCmd) {
            commInput.removeListener(this);
            commInput.close();
            removeCommand(closeCommCmd);
            addCommand(openCommCmd);
        } else if (c == clearCmd) {
            setString(null);
        }
    }
    
    /**
     * Called in order to make the current string value such that it can
     * be later selected from a list of previous values.<p>
     * 
     * If selection wasn't enabled in the constructor call then this method
     * call will be ignored.<p>
     */
    public void storeValue() {
        
        if (selectable) {
            setPreviousValues(StringValuesManager.addPreviousValue(getId(), getString(), ignoreCase));   
        }
    }
    
    /**
     * List class used for selecting one of the previous values.
     */
    private class PreviousValueSelection extends List implements CommandListener {

        private Command backCmd = new Command("Back", Command.BACK, 1);
        
        private Command selectCmd = new Command("Select", Command.OK, 1);
        
        private Command deleteCmd = new Command("Delete", Command.OK, 2);
        
        private Command deleteAllCmd = new Command("Delete all", Command.OK, 3);
        
        
        PreviousValueSelection(String[] prevValues) {
            super("Previous values", List.IMPLICIT);
            setCommandListener(this);
            addCommand(backCmd);
            setSelectCommand(selectCmd);
            initList(prevValues);
        }
        
        private void initList(String[] values) {
            removeCommand(deleteCmd);
            removeCommand(deleteAllCmd);
            deleteAll();
            
            if (values != null) {
                
                for (int i = 0; i < values.length; i++) {
                    append(values[i], null);
                    
                    if (selectionFont != null) {
                        setFont(i, selectionFont);
                    }
                }     
                
                if (values.length > 0) {
                    addCommand(deleteCmd);
                    addCommand(deleteAllCmd);
                }
            }
        }
        
        /**
         * @see javax.microedition.lcdui.CommandListener#commandAction(javax.microedition.lcdui.Command, javax.microedition.lcdui.Displayable)
         */
        public void commandAction(Command c, Displayable d) {
            
            if (c == backCmd) {
                ViewHandler.getInstance().displayPrevious();
            } else if (c == selectCmd) {
                setString(getString(getSelectedIndex()));
                ViewHandler.getInstance().displayPrevious();
            } else if (c == deleteCmd) {
                String[] newSet = StringValuesManager.removeGroupValue(getId(), getString(getSelectedIndex()));
                
                if (newSet != null) {
                    setPreviousValues(newSet);
                    initList(previousValues);    
                }
                
            } else if (c == deleteAllCmd) {
                String[] newSet = StringValuesManager.removeAllGroupValues(getId());
                setPreviousValues(newSet);
                initList(previousValues);
            }
        }
    }
    
}
