package com.nokia.nfc.app.test.manual.ui;

import javax.microedition.lcdui.*;

import com.nokia.nfc.app.util.Log;
import com.nokia.nfc.app.util.LogLevel;

/**
 * Abstract base class for test views. All {@link TestForm} classes must 
 * provide a default constructor.<p>
 * 
 * When the test is opened by the user by selecting it from the list of available
 * tests the method {@link #open()} will be called before setting the test
 * view visible. When user returns back to the list from the test view the
 * method {@link #close()} will be called.<p>
 */
public abstract class TestForm extends Form implements CommandListener, ItemCommandListener {

    private Command backCmd = new Command("Back", Command.BACK, 1);
    
    private Command viewLogCmd = new Command("View log", Command.OK, 100);
    
    
    /**
     * @param title title of the form.
     * @param addDefaultCommands if <code>false</code> then the default commands
     *                           will not be added to the form.
     */
    protected TestForm(String title, boolean addDefaultCommands) {
        super(title);
        setCommandListener(this);
        
        if (addDefaultCommands) {
            addDefaultCommands();
        }
    }

    protected final void log(String msg) {
        if (LogLevel.level >= Log.APPLICATION) Log.log(msg);
    }
    
    protected final void log(String msg, boolean value) {
        if (LogLevel.level >= Log.APPLICATION) Log.log(msg, value);
    }

    protected final void log(String msg, byte[] data) {
        if (LogLevel.level >= Log.APPLICATION) Log.log(msg, data);
    }
    
    protected final void log(String msg, Object obj) {
        if (LogLevel.level >= Log.APPLICATION) Log.log(msg, obj);
    }
    
    protected final void log(String msg, int value) {
        if (LogLevel.level >= Log.APPLICATION) Log.log(msg, value);
    }
    
    protected final void displayResult(String resultMsg, boolean error) {
        
        if (resultMsg != null) {
            log(resultMsg);
            
            if (error) {
                ViewHandler.getInstance().displayTimeoutAlert(resultMsg, AlertType.ERROR, Alert.FOREVER);    
            } else {
                ViewHandler.getInstance().displayTimeoutAlert(resultMsg, AlertType.INFO, 1000);
            }
        }
    }
    
    /**
     * @param title title of the form.
     */
    protected TestForm(String title) {
        this(title, true);
    }
    
    /**
     * Called when the test is opened. If time consuming operations need to
     * be done then one ought to do it in a separate thread. 
     */
    public void open() { 
    }
    
    /**
     * Called when the test is closed. Clean up resources in this method
     * if necessary.<p>
     */
    public void close() {
        
        for (int i = 0; i < size(); i++) {
            Item item = get(i);
            
            if (item instanceof SelectableTextField) {
                ((SelectableTextField) item).cleanUp();
            }
        }
    }
    
    /**
     * Add the default commands to the form.
     */
    private void addDefaultCommands() {
        addCommand(backCmd);
        addCommand(viewLogCmd);
    }
    
    /**
     * Adds a new command to the form. Type of the command will be {@link Command#OK}
     * and the label is defined by the parameter <code>label</code>. These commands
     * can be handled by overriding the {@link #handleOkCmd(String)} and comparing
     * that method's parameter to the label passed to this method. This is intended
     * to make adding new commands a bit easier but sub classes can also add
     * commands normally if they wish.
     * 
     * @param name label of the command.
     */
    protected void addOkCmd(String label) {
        addCommand(new Command(label, Command.OK, 1));
    }
    
    protected void addOkButton(String label) {
        StringItem button = new StringItem(null, label, StringItem.BUTTON);
        button.setItemCommandListener(this);
        button.addCommand(new Command(label, Command.ITEM, 1));
        append(button);
    }
    
    public void commandAction(Command c, Item item) {
        startOkCmdHandling(c.getLabel());
    }
    
    /**
     * Handles commands common to all TestForm classes. Subclasses can override
     * this method but then they should call <code>super.commandAction()</code>
     * so that the default commands get handled.
     * 
     * @see javax.microedition.lcdui.CommandListener#commandAction(javax.microedition.lcdui.Command, javax.microedition.lcdui.Displayable)
     */
    public void commandAction(Command c, Displayable d) {
        
        if (c == backCmd) {
            ViewHandler.getInstance().displayPrevious();
        } else if (c == viewLogCmd) {
            ViewHandler.getInstance().display(new LogView());
        } else if (c.getCommandType() == Command.OK) {
            startOkCmdHandling(c.getLabel());
        } 
    }
    
    private void startOkCmdHandling(final String label) {
        new Thread() {
            public void run() {
                handleOkCmd(label);
            }
        }.start();
    }
    
    /**
     * Handles non-default commands of type {@link Command#OK} - most likely
     * commands that have been added using the method {@link #addOkCmd(String)}.<p>
     * 
     * Sub classes that use {@link #addOkCmd(String)} must override this method.
     * By default this method implementation does nothing.<p>
     * 
     * A separate thread will be automatically created for this method to execute in.
     * This way the sub class can perform time consuming operations in the command
     * handler.<p>
     * 
     * @param label label of the command selected by the user.
     */
    protected void handleOkCmd(String label) {
    }
    
}
