package com.nokia.nfc.app.test.manual.ui;

import java.util.Stack;

import javax.microedition.lcdui.*;

import com.nokia.nfc.app.test.manual.Control;

/**
 * This class handles displaying views.<p>
 * 
 * During the application startup the method {@link ViewHandler#initialize(Display, Control)}
 * must be called before trying to acquire a {@link ViewHandler} instance.<p>
 */
public class ViewHandler implements CommandListener {

    public static final int DEFAULT_ALERT_TIMEOUT = 3000;
    
    private static ViewHandler instance;
    
    private Command dismissAlertCmd = new Command("OK", Command.OK, 1);
    
    private Display display;
    
    private Control ctrl;
    
    private Stack views;
    
    
    public synchronized static void initialize(Display display, Control ctrl) {
        instance = new ViewHandler(display, ctrl);    
    }
    
    public synchronized static ViewHandler getInstance() {
        
        if (instance == null) {
            throw new IllegalStateException("initialize() must be called first");
        }
        
        return instance;
    }
    
    public synchronized static void cleanUp() {
        instance = null;
    }
    
    private ViewHandler(Display display, Control ctrl) {
        this.display = display;
        this.ctrl = ctrl;
        this.views = new Stack();
    }
    
    /**
     * @see javax.microedition.lcdui.CommandListener#commandAction(javax.microedition.lcdui.Command, javax.microedition.lcdui.Displayable)
     */
    public void commandAction(Command c, Displayable d) {
        
        if (c == dismissAlertCmd) {
            displayPrevious();
        }
    }
    
    public synchronized void setCurrentItem(Item item) {
        display.setCurrentItem(item);
    }
    
    public void displayTimeoutAlert(String text, AlertType type) {
        displayTimeoutAlert(text, type, DEFAULT_ALERT_TIMEOUT);
    }
    
    public void displayTimeoutAlert(String text, AlertType type, int timeout) {
        Alert alert = new Alert(null, text, null, type);
        alert.setTimeout(timeout);
        alert.addCommand(dismissAlertCmd);
        alert.setCommandListener(this);
        displayAlert(alert);
    }

    public void displayAlert(Alert alert) {
        display(alert, true);
    }
    
    /**
     * Displays the specified view. Display alerts with the methods meant for alerts.
     */
    public synchronized void display(Displayable view) {
        
        if (view instanceof Alert) {
            throw new IllegalStateException("Not for alerts this is.");
        }
        
        display(view, true);
    }
    
    private synchronized void display(Displayable view, boolean push) {
        
        if (view instanceof Alert) {
            // Display.setCurrent(Displayable) would use Display.getCurrent() as the next view.
            Alert alert = (Alert) view;
            Displayable nextView = null;
            
            // Look for the next non Alert view. Otherwise an IllegalStateException
            // would be thrown if trying to call Display.setCurrent(Alert, Alert).
            // The reasoning here is that if there would be a case where two
            // alerts are on top of the view stack then only one of those would
            // be something like a question alert that shouldn't get lost.
            // The other would be an alert with a timeout -> ignorable.
            for (int i = (views.size() - 2); i >= 0; i--) {
                
                if (!(views.elementAt(i) instanceof Alert)) {
                    nextView = (Displayable) views.elementAt(i);
                }
            }
            
            if (nextView == null) {
                // Otherwise an NullPointerException would be thrown.
                // If this scenario ever occurs, then it is assumed that the alert
                // has a CommandListener that handles closing the alert and
                // displays something meaningful.
                // The default alert dismiss command would cause the application to exit.
                display.setCurrent(alert, new Form(""));
            } else {
                display.setCurrent(alert, nextView);    
            }
            
        } else {
            display.setCurrent(view);    
        }
        
        if (push) {
            views.push(view);
        }
    }
    
    public synchronized void displayPrevious() {
        views.pop();
        
        if (views.size() == 0) {
            // No more views to display -> exit application.
            ctrl.exit();
        } else {
            Displayable previous = (Displayable) views.peek();
            display(previous, false);
        }
    }
    
}
