package com.nokia.nfc.app.util;

/**
 * Class that maintains a fixed size string buffer. Provides a method to get
 * the last N strings from the buffer.
 */
public class BackBuffer {

    private final int backBufferLen;
    
    private final String[] backBuffer;
    
    private int backBufferLatest = -1;

    
    public BackBuffer(int len) {
        backBufferLen = len;
        backBuffer = new String[backBufferLen];
    }
    
    public synchronized void clear() {
        
        for (int i = 0; i < backBuffer.length; i++) {
            backBuffer[i] = null;
        }
        
        backBufferLatest = -1;
    }
    
    public synchronized void addToBackBuffer(String line) {
        backBufferLatest++;
        
        if (backBufferLatest >= backBufferLen) {
            backBufferLatest = 0;
        } 

        backBuffer[backBufferLatest] = line;
    }
    
    /**
     * Return last N lines of the log. N is specified by parameter <code>lines</code>.
     * Note that some lines can be <code>null</code> if there aren't as many
     * lines in the log's back buffer. If there are <code>null</code> values
     * then those are at the beginning of the returned array.<p>
     * 
     * If the <code>lines</code> value is greater than buffer length then
     * the value is changed to be equal to buffer length.<p>
     * 
     * If the <code>lines</code> amount is <code>-1</code> then 
     * the value is changed to be equal to buffer length.<p>
     * 
     * @param lines amount of lines to return.
     * @return last N lines.
     */
    public synchronized String[] tail(int lines) {
        
        if (lines > backBufferLen || lines == -1) {
            lines = backBufferLen;
        }
        
        String[] ret = new String[lines];
        
        if (backBufferLatest >= 0) {
            int secondPartLen = backBufferLatest + 1;
            
            if (secondPartLen - lines < 0) {
                int firstPartLen = lines - secondPartLen;
                System.arraycopy(backBuffer, backBufferLen - firstPartLen, ret, 0, firstPartLen);
                System.arraycopy(backBuffer, 0, ret, firstPartLen, secondPartLen);
            } else {
                System.arraycopy(backBuffer, secondPartLen - lines, ret, 0, lines);
            }
        }
        // Otherwise all elements are null -> no copying needed
        
        return ret;
    }

}
