package com.nokia.nfc.app.util;

import java.io.*;
import java.util.Vector;

import javax.microedition.io.CommConnection;
import javax.microedition.io.Connector;
import javax.microedition.lcdui.TextField;

/**
 * Class that can be used for listening input from a COMM port. This way
 * the user can type the text on a PC terminal and then press enter and
 * the text will be copied to a {@link TextField} for example. This
 * class reads a line at a time. Any other characters read after a new line or 
 * carriage return character are ignored.<p>
 * 
 * Note: it seems that <code>System.out</code> calls do not send the data to the PC. 
 * Instead one should use the {@link OutputStream} associated with 
 * the {@link CommConnection}. It is recommended that {@link #close()} is
 * called once input doesn't need to be read any more.<p>
 * 
 * Non ASCII characters will likely not be copied correctly. This might
 * be more related to the terminal program than this class. This class
 * uses UTF-8 encoding to read from the input stream.<p>
 */
public class CommPortInput implements Runnable {

    private static final String COMM_PORT_PROPERTY = "microedition.commports";

    private static final String COMM_PORT_URL_PREFIX = "comm:";
    
    private static final char COMM_PORT_SEPARATOR = ',';
    
    private String uri;

    private InputStreamReader isr;
    
    private CommConnection conn;
    
    private Vector listeners;
    
    
    public CommPortInput(String uri) {
        this.uri = uri;
        this.listeners = new Vector();
    }
    
    public static CommPortInput createDefaultInput() {
        return new CommPortInput(getFirstCommPortURI());
    }

    public synchronized boolean open() {
        
        if (conn == null) {
            
            try {
                if (LogLevel.level >= Log.ERROR) Log.log("Open CommConnection", uri);
                conn = (CommConnection) Connector.open(uri);
                isr = new InputStreamReader(conn.openInputStream(), "UTF-8");
                // ^ For some reason only 'UTF-8' works (nothing, 'US-ASCII', US_ASCII' were tried too).
                // Otherwise no input is read. Used TeraTerm 3.1.3.

                new Thread(this).start();
                return true;
            } catch (Exception ioe) {
                close();
                if (LogLevel.level >= Log.ERROR) Log.log("Failed to open COMM port input", ioe);
            }
        }
        // Otherwise already open.
        
        return false;
    }
    
    public synchronized void close() {
        
        if (isr != null) {
            try {
                isr.close();
            } catch (IOException e) {
            } finally {
                isr = null;
            }
        }
        
        if (conn != null) {
            try {
                conn.close();
            } catch (IOException e) {
            } finally {
                conn = null;
            }
        }
    }
    
    public synchronized void addListener(InputListener listener) {
        
        if (listeners.indexOf(listener) == -1) {
            listeners.addElement(listener);   
        }
    }
    
    public synchronized void removeListener(InputListener listener) {
        listeners.removeElement(listener);
    }
    
    private synchronized void notifyListeners(String input) {
        InputListener il = null;
        
        for (int i = 0; i < listeners.size(); i++) {
            il = (InputListener) listeners.elementAt(i);
            il.notifyInput(input, this);
        }
    }
        
    private static final String getFirstCommPortURI() {
        String ports = System.getProperty(COMM_PORT_PROPERTY);
        int separatorIndex = ports.indexOf(COMM_PORT_SEPARATOR);
        String portId = null;
        
        if (separatorIndex != -1) {
            portId = ports.substring(0, separatorIndex);
        } else {
            portId = ports;
        }
        
        return COMM_PORT_URL_PREFIX + portId;
    }
    
    /**
     * Thread that is used for reading input. Always reads as many bytes
     * are available in the stream.
     * 
     * @see java.lang.Runnable#run()
     */
    public void run() {
        String line = null;
        
        while (isr != null) {
            line = readLine();
            
            if (line != null) {
                notifyListeners(line);
            }
        }
    }
    
    /**
     * Reads until EOS or linefeed character and discards the rest.
     */
    private final String readLine() {
        
        try {
            char[] buf = new char[1024];
            int len = -1;
            boolean end = false;
            StringBuffer dst = new StringBuffer(); 
            
            while (!end && (len = isr.read(buf)) != -1) {

                for (int i = 0; i < len; i++) {

                    switch (buf[i]) {
                        case '\n':
                        case '\r':
                            end = true;
                            len = i;
                            break;
                    }
                }

                dst.append(buf, 0, len);
            }

            return dst.toString();
        } catch (Exception ioe) {
            return null;
        }
    }
    
}
