package com.nokia.nfc.app.util;

import java.io.IOException;
import java.io.PrintStream;
import java.util.Calendar;
import java.util.Vector;

import javax.microedition.io.Connector;
import javax.microedition.io.file.FileConnection;

/**
 * Simple class for logging messages to either console and/or files.
 * Support for log levels isn't provided but instead two constants
 * {@link #TEST} and {@link #IMPL} are provided. In the source code
 * one should use one of these constants to check whether logging should
 * be done.
 * <pre>
 *   // In test code:
 *   if (Log.TEST) Log.note("Test started");
 *   
 *   // In stub implementation:
 *   if (Log.IMPL) Log.note("Method x called");
 * </pre>
 * <p>
 * 
 * For log files there are two locations to which the Log class tries
 * to create the file to. First the memorycard is tried (<code>fileconn.dir.memorycard</code>), 
 * and if that wasn't successful then the Recordings directory is tried
 * (<code>fileconn.dir.recordings</code>).
 * <p>
 * 
 * The implementation assumes that the FileConnection API is supported.
 * Otherwise a {@link ClassNotFoundException} will be thrown when trying
 * to access this class.
 * <p>
 */
public class Log {
    
    /**
     * DO_DOCUMENT
     */
    public static final int NONE = 0;
    
    /**
     * DO_DOCUMENT
     */
    public static final int APPLICATION = 1;
    
    /**
     * DO_DOCUMENT
     */
    public static final int ERROR = 2;
    
    /**
     * DO_DOCUMENT
     */
    public static final int WARNING = 3;
    
    /**
     * DO_DOCUMENT
     */
    public static final int INFO = 4;
    
    /**
     * DO_DOCUMENT
     */
    public static final int DEBUG = 5;

    
    /** Index of the {@link PrintStream} attached to a file in the {@link #streams} array. */
    private static final int FILE_INDX = 0;
    
    /** Index of the {@link PrintStream} attached to a <code>System.out</code> in the {@link #streams} array. */
    private static final int CONSOLE_INDX = 1;
    
    /** 
     * Holds the {@link PrintStream} objects which are used for outputting log messages.
     * If an array element value is <code>null</code> then it is skipped.
     */
    private static PrintStream[] streams = new PrintStream[2];
    
    /** If not <code>null</code> then this is the {@link FileConnection} that is being used for log output. */
    private static FileConnection fileConn = null;
    
    /** Registered {@link LogListener} objects.  */
    private static Vector listeners = new Vector();
    
    /** Length of the log back buffer, i.e. how many last lines to keep in memory. */
    private static final int BACK_BUFFER_LEN = 100;
    
    /** The log back buffer. */
    private static final BackBuffer BACK_BUFFER = new BackBuffer(BACK_BUFFER_LEN);
    
    
    /**
     * Enable or disable printing log messages to a file.
     * File logging should be disabled when exiting the MIDlet as it
     * will also close the opened {@link FileConnection}.
     * 
     * @param enabled if <code>true</code> then a new file is created, 
     *                <code>false</code> disables logging to a file.
     * @param filenamePrefix prefix of the filename part of the file path.
     * 
     * @throws IOException if a file could not be created.
     * @throws SecurityException if no permission was given to create a file.
     */
    public static void setFileLogger(boolean enabled, String filenamePrefix) 
        throws IOException, SecurityException 
    {        
        if (enabled) {
            openFileConnection(filenamePrefix);
        } else {
            
            if (streams[FILE_INDX] != null) {
                streams[FILE_INDX].close();
                streams[FILE_INDX] = null;
            }
            
            if (fileConn != null) {
                
                try {
                    fileConn.close();
                } catch (IOException e) {
                }
                
                fileConn = null;
            }
        }
    }
    
    /**
     * Opens the {@link FileConnection} and a {@link PrintStream}.
     * 
     * @param filenamePrefix prefix of the filename part of the file path.
     * 
     * @throws IOException if a file could not be created.
     * @throws SecurityException if no permission was given to create a file.
     * @throws NullPointerException if <code>filenamePrefix</code> is <code>null</code>.
     */
    private static void openFileConnection(String filenamePrefix) 
        throws IOException, SecurityException 
    {
        if (!System.getProperty("microedition.io.file.FileConnection.version").equals("1.0")) {
            // FileConnection API version 1.0 isn't supported.
            // Probably a bit unnecessary check as if it isn't supported
            // a ClassNotFoundException would have been thrown earlier.
            throw new IOException("FileConnection not available");
        }
        
        final String filename = createLogFilename(filenamePrefix);
        final String[] pathProperties = {"fileconn.dir.memorycard", "fileconn.dir.recordings"};
        String path = null;
        
        // Attempt to create a file to the directories specified by the 
        // system properties in array pathProperties.
        for (int i = 0; i < pathProperties.length; i++) {
            path = System.getProperty(pathProperties[i]);
            
            // Only throw declared exceptions if this is the last path
            // to try.
            try {
                
                if (path == null) {
                    
                    if (i < (pathProperties.length - 1)) {
                        continue;
                    } else {
                        throw new IOException("Path not available: " + pathProperties[i]);
                    }
                }

                FileConnection fConn = (FileConnection) 
                    Connector.open(path + filename, Connector.WRITE);
                
                // Assume that createLogFilename creates such a filename
                // that is enough to separate filenames even if they
                // are created in a short interval (seconds).
                fConn.create();
                streams[FILE_INDX] = new PrintStream(fConn.openDataOutputStream());
                
                // Opening the connection and stream was succesful so don't
                // try other paths.
                fileConn = fConn;
                break;
            } catch (SecurityException se) {
                if (i == (pathProperties.length - 1)) {
                    throw se;
                }
            } catch (IOException ioe) {
                if (i == (pathProperties.length - 1)) {
                    throw ioe;
                }                
            }
        }
    }
    
    /**
     * Creates a filename to be used for the log file.
     * 
     * @param filenamePrefix prefix for the filename.
     * @return a log file name.
     */
    private static String createLogFilename(String filenamePrefix) {
        StringBuffer fn = new StringBuffer(filenamePrefix.length() + 1 + 17 + 4);
        fn.append(filenamePrefix);
        fn.append('_');
        appendDateAndTime(fn);
        fn.append(".log");
        return fn.toString();
    }

    /**
     * Appends 2 digits to the given {@link StringBuffer}. If the 
     * <code>num</code> is less than 10 a 0 will be appended before it. 
     * 
     * @param sb the destination for the digits.
     * @param num the positive integer number to append.
     * 
     * @throws NullPointerException if <code>sb</code> is <code>null</code>.
     */
    private static final void append2Digits(StringBuffer sb, int num) {
        
        if (num < 10) {
            sb.append('0');
        }
        
        sb.append(num);
    }
    
    /**
     * Append a timestamp to the given {@link StringBuffer}. 
     * The syntax is <code>ddMMyyyy_hhMMss</code>
     * for date. Example: <code>26042007_102404</code>.
     * 
     * @param sb the destination for the timestamp.
     * 
     * @throws NullPointerException if <code>sb</code> is <code>null</code>.
     */
    public static void appendDateAndTime(StringBuffer sb) {
        Calendar c = Calendar.getInstance();

        int day = c.get(Calendar.DAY_OF_MONTH);
        int month = c.get(Calendar.MONTH) + 1;
        int year = c.get(Calendar.YEAR);
        append2Digits(sb, day);
        append2Digits(sb, month);
        sb.append(year);
        sb.append('_');
        
        int hour = c.get(Calendar.HOUR_OF_DAY);
        int mins = c.get(Calendar.MINUTE);
        int secs = c.get(Calendar.SECOND);
        append2Digits(sb, hour);
        append2Digits(sb, mins);
        append2Digits(sb, secs);
    }
    
    /**
     * Converts given time in milliseconds to HHMMSS.mmm format and
     * appends the generated string to StringBuffer.
     * 
     * @param time time in milliseconds.
     * @return the given time as a string.
     */
    private static final void appendLogTime(long time, StringBuffer dst) {
        long msecs, secs, mins, hours;
        time -= msecs = time % 1000L;
        time -= secs = time % (60L * 1000L) / 1000L;
        time -= mins = time % (60L * 60L * 1000L) / (60L * 1000L);
        hours = time % (24L * 60L * 60000L) / (60L * 60L * 1000L);
        time = hours * 10000L + mins * 100L + secs;

        for (long l = 100000L; l > time && l > 1L; l /= 10)
            dst.append("0");
        dst.append(time).append(".");
        for (long l = 100L; l > msecs && l > 1L; l /= 10)
            dst.append("0");
        dst.append(msecs);
    }
    
    /**
     * Enable or disable printing log messages to console (<code>System.out</code>).
     * 
     * @param enabled defines whether console log is enabled (<code>true</code>) or disabled (<code>false</code>).
     */
    public static void setConsoleLogger(boolean enabled) {
        
        if (enabled) {
            streams[CONSOLE_INDX] = System.out;
        } else {
            streams[CONSOLE_INDX] = null;
        }
    }
    
    /**
     * @return <code>true</code> if printing log output to a file is enabled.
     */
    public static boolean isUsingFiles() {
        return (streams[FILE_INDX] != null);
    }
    
    /**
     * @return <code>true</code> if printing log output to console is enabled.
     */
    public static boolean isUsingConsole() {
        return (streams[CONSOLE_INDX] != null);
    }
    
    /**
     * Print the specified message to the enabled streams. A <code>null</code>
     * message will be replaced with an empty string.
     * 
     * @param printObj if <code>true</code> then then the <code>obj</code> Object will be included in the message.
     * @param message the message to print.
     * @param obj the optional object to include in the message.
     */
    private static void print(boolean printObj, String message, Object obj) {
        String objStr = (obj == null ? "NULL" : obj.toString());
        
        if (message == null) {
            message = "";
        }
        
        // Print the line with a single println to avoid possible
        // linebreaks when printing to console.
        StringBuffer sb = new StringBuffer(
                1 + message.length() + (printObj ? objStr.length() + 1 : 0));
        appendLogTime(System.currentTimeMillis(), sb);
        sb.append(' ');
        sb.append(message);
        
        if (printObj) {
            sb.append(": ");
            sb.append(objStr);
        }
        
        String line = sb.toString();
        
        for (int i = 0; i < streams.length; i++) {
            PrintStream ps = streams[i];
            
            if (ps != null) {
                ps.println(line);
                ps.flush();
            }
        }
        
        BACK_BUFFER.addToBackBuffer(line);
        notifyListeners(line);
    }
    
    /**
     * Prints the specified message to enabled log streams.
     * 
     * @param message message to print. If null only timestamp will be printed.
     */
    public static void log(String message) {
        print(false, message, null);
    }
    
    /**
     * Prints the specified message and the object to enabled log streams.
     * 
     * @param message message to print. If null only timestamp and value will be printed.
     */
    public static void log(String message, Object value) {
        print(true, message, value);
    }
    
    /**
     * Prints the specified message and integer value to enabled log streams.
     * 
     * @param message message to print. If null only timestamp and value will be printed.
     */
    public static void log(String message, int value) {
        print(true, message, new Integer(value));
    }
    
    /**
     * Prints the specified message and long integer value to enabled log streams.
     * 
     * @param message message to print. If null only timestamp and value will be printed.
     */
    public static void log(String message, long value) {
        print(true, message, new Long(value));
    }
    
    /**
     * Prints the specified error message and boolean value to enabled log streams.
     * 
     * @param message message to print. If null only timestamp and value will be printed.
     */
    public static void log(String message, boolean value) {
        print(true, message, new Boolean(value));
    }

    /**
     * Prints the specified error message and boolean value to enabled log streams.
     * 
     * @param message message to print. If null only timestamp and value will be printed.
     */
    public static void log(String message, byte[] data, int len) {
        print(true, message, bytes2hex(data, 0, len, true));
    }
    
    /**
     * Prints the specified error message and boolean value to enabled log streams.
     * 
     * @param message message to print. If null only timestamp and value will be printed.
     */
    public static void log(String message, byte[] data) {
        String dataStr = (data == null ? null : bytes2hex(data, 0, data.length, true));
        print(true, message, dataStr);
    }

    /**
     * Returns the given byte array as a hex string. Starts from the specified
     * offset and and reads up to <code>len</code> bytes.
     */
    public static final String bytes2hex(byte[] data, int offset, int len, boolean limit) {
        
        if (data == null) {
            return "";
        }
        
        if (LogLevel.MAX_HEX_LEN != -1 && len > LogLevel.MAX_HEX_LEN) {
            len = LogLevel.MAX_HEX_LEN;
        } else {
            limit = false;
        }
        
        StringBuffer sb = new StringBuffer(2 * len + (limit ? 3 : 0));
        int lastIndx = offset + len;
        
        for (int i = offset; i < lastIndx; i++) {
            int b = data[i] & 0xFF;
            
            if (b < 0x10) {
                sb.append('0');
            } 

            sb.append(Integer.toHexString(b));
        }
        
        if (limit) {
            sb.append("...");
        }
        
        return sb.toString();
    }
    
    private static void notifyListeners(String line) {
        
        synchronized (listeners) {
            
            for (int i = 0; i < listeners.size(); i++) {
                ((LogListener) listeners.elementAt(i)).lineAdded(line);
            }
        }
    }
    
    public static void addListener(LogListener listener) {

        synchronized (listeners) {

            if (listeners.indexOf(listener) == -1) {
                listeners.addElement(listener);    
            }
        }
    }

    public static void removeListener(LogListener listener) {
        
        synchronized (listeners) {
            listeners.removeElement(listener);
        }
    }
    
    public static BackBuffer getBackBuffer() {
        return BACK_BUFFER;
    }
    
}
