package com.nokia.nfc.app.util;

import java.io.*;
import java.util.Enumeration;
import java.util.Hashtable;

import javax.microedition.midlet.MIDlet;
import javax.microedition.rms.*;


public class Settings {
    
    /** Name of the record store where settings are stored. Only the first record is used. */
    private static final String STORE_NAME = "settings";
    
    private static Hashtable settings;
    
    private static MIDlet midlet;
    
    
    public static synchronized void initialize(MIDlet midlet) {
        Settings.midlet = midlet;
        load();
    }
    
    /**
     * Set the value of a setting and save the settings to RMS.
     * 
     * @param name name of the setting.
     * @param value new value for the setting, must not be <code>null</code>.
     */
    public static synchronized void setSetting(String name, String value) {
        
        if (value == null) {
            throw new IllegalArgumentException("Null setting value");
        }
        
        settings.put(name, value);
        save();
    }
    
    /**
     * Return the value of a setting.
     * 
     * @param name name of the setting, must not be <code>null</code>.
     * @param allowJad if <code>true</code> then look for a value from JAD file
     *                 if the setting hasn't been saved to RMS.
     * @return the setting value.
     */
    public static synchronized String getSetting(String name, boolean allowJad) {
        String value = (String) settings.get(name);
        
        if (value == null && allowJad) {
            value = midlet.getAppProperty(name);
        }
        
        return value;
    }
    
    public static int getIntSetting(String name, int defaultValue) {
        String strValue = getSetting(name, false);
        
        if (strValue != null) {
            
            try {
                return Integer.parseInt(strValue);
            } catch (NumberFormatException e) {
            }
        }

        return defaultValue;
    }
    
    public static void setIntSetting(String name, int value) {
        setSetting(name, Integer.toString(value));
    }
    
    private static void load() {
        settings = new Hashtable();
        RecordStore rs = null;
        
        try {
            rs = RecordStore.openRecordStore(STORE_NAME, true);
            byte[] recordData = null;
            
            if (rs.getNumRecords() == 0) {
                // No settings have been saved yet.
                return;
            } else {
                recordData = rs.getRecord(1);
            }
            
            ByteArrayInputStream bais = new ByteArrayInputStream(recordData);
            DataInputStream dis = new DataInputStream(bais);
            int pairs = dis.readInt();
            
            while (pairs > 0) {
                String key = dis.readUTF();
                String value = dis.readUTF();
                settings.put(key, value);
                pairs--;
            }
            
            // Assume that writing newer fails and no corruption happens.
            // Otherwise we'd better create a new record first and then
            // delete the previous one.
            
            // Assume that all settings always fit to a single record.
            // Should be a valid assumption if Settings is only used for
            // saving a few settings: 256kB record would fit ~1600 settings
            // where each name and value is 20 characters long and each
            // character consumes maximum 4 bytes.
        } catch (Exception e) {
            if (LogLevel.level >= Log.ERROR) Log.log("Failed to save settings to RMS", e);
        } finally {
            
            if (rs != null) {
                try {
                    rs.closeRecordStore();
                } catch (Exception e) {
                }
            }
        }
    }
    
    private static void save() {
        RecordStore rs = null;
        
        try {
            rs = RecordStore.openRecordStore(STORE_NAME, true);
            
            if (settings == null) {
                throw new IllegalStateException("initialize() hasn't been called");
            }
            
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            DataOutputStream dos = new DataOutputStream(baos);
            dos.writeInt(settings.size());
            Enumeration keys = settings.keys();
            
            while (keys.hasMoreElements()) {
                String key = (String) keys.nextElement();
                String value = (String) settings.get(key);
                dos.writeUTF(key);
                dos.writeUTF(value);
                // setSetting() guarantees that there are no null values and
                // keys cannot be null.
            }
            
            dos.flush();
            byte[] recordData = baos.toByteArray();
            
            // Reuse the first record if it exists.
            if (rs.getNextRecordID() == 1) {
                rs.addRecord(recordData, 0, recordData.length);
            } else {
                rs.setRecord(1, recordData, 0, recordData.length);
            }
            
            // Assume that writing newer fails and no corruption happens.
            // Otherwise we'd better create a new record first and then
            // delete the previous one.
            
            // Assume that all settings always fit to a single record.
            // Should be a valid assumption if Settings is only used for
            // saving a few settings: 256kB record would fit ~1600 settings
            // where each name and value is 20 characters long and each
            // character consumes maximum 4 bytes.
        } catch (Exception e) {
            if (LogLevel.level >= Log.ERROR) Log.log("Failed to save settings to RMS", e);
        } finally {
            
            if (rs != null) {
                try {
                    rs.closeRecordStore();
                } catch (Exception e) {
                }
            }
        }
    }

}
