package com.nokia.nfc.app.util;

import java.io.*;
import java.util.Hashtable;

import javax.microedition.io.Connector;
import javax.microedition.io.file.FileConnection;


/**
 * Class for loading and saving string values. String values are grouped based
 * on group identifiers. 
 */
public class StringValuesManager {
    
    /** Extension used for string values files. */
    private static final String EXTENSION = ".str_dat";
    
    /** 
     * List of system properties defining possible directories for storing 
     * string values. The value is attempted first. 
     */
    private static final String[] LOCATION_PROPS = {
        "fileconn.dir.memorycard",
        "fileconn.dir.recordings"
    }; 
    
    /** Keys are identifier strings and values are string arrays. */
    private static Hashtable cachedGroups = new Hashtable();
    
    
    /**
     * Removes all values from the specified group.
     */
    public synchronized static String[] removeAllGroupValues(String id) {
        FileConnection conn = openFile(id, Connector.READ_WRITE);
        
        if (conn != null) {
            try {
                saveGroupValues(null, conn);
            } finally {
                try {
                    conn.close();
                } catch (IOException e) {
                }
            }
        }

        cachedGroups.remove(id);
        return null;
    }

    /**
     * Removes a single value from a group. The group values must have been
     * loaded before calling this method.
     * 
     * @return the resulting group array or <code>null</code> if the specified value wasn't found.
     */
    public synchronized static String[] removeGroupValue(String id, String value) {
        String[] values = (String[]) cachedGroups.get(id);
        
        if (values != null) {
            String[] newSet = removeValue(values, value);
            
            if (newSet != null) {
                
                if (newSet.length == 0) {
                    // Resulting set is empty -> remove file altogether. 
                    removeAllGroupValues(id);
                } else {
                    // Removed the specified value.
                    FileConnection conn = openFile(id, Connector.READ_WRITE);
                    
                    if (conn != null) {
                        
                        try {
                            saveGroupValues(newSet, conn);    
                        } finally {
                            try {
                                conn.close();
                            } catch (IOException e) {
                            }
                        }
                    }
                }
                
                cachedGroups.put(id, newSet);
                return newSet;
            }
        }
        // Otherwise nothing to remove from.

        return null;
    }
    
    private static final String[] removeValue(String[] set, String value) {
        int index = -1;
        
        for (int i = 0; i < set.length; i++) {
            
            if (set[i].equals(value)) {
                index = i;
                break;
            }
        }

        if (index != -1) {
        
            if (set.length > 1) {
                System.out.println("remove from index " + index);
                int dstOffset = 0;
                int beforeLen = index;
                int afterLen = (set.length - index - 1);
                
                String[] ret = new String[set.length - 1];
                
                if (beforeLen > 0) {
                    System.arraycopy(set, 0, ret, 0, beforeLen);
                    dstOffset = index;
                }
                
                if (afterLen > 0) {
                    System.arraycopy(set, index + 1, ret, dstOffset, afterLen);
                }
                
                return ret;
            } else {
                return new String[0];
            }

        } else {
            return null;
        }
    }
    
    /**
     * Load string values from a file.
     * 
     * @param id identifier for the group of string values.
     * @return the string values or <code>null</code> if none were available.
     */
    public synchronized static String[] loadGroupValues(String id, String[] initialValues) {
        String[] cached = (String[]) cachedGroups.get(id);
        
        if (cached == null) {
            FileConnection conn = openFile(id, Connector.READ_WRITE);
            
            if (!conn.exists()) {
                // Store 'initialValues' to the file.
                saveGroupValues(initialValues, conn);
                cached = initialValues;
                
            } else if (conn != null) {
                
                try {
                    cached = loadGroupValues(id, conn);   
                } finally {
                    try {
                        conn.close();
                    } catch (IOException e) {
                    }
                }
            }

            if (cached != null) {
                cachedGroups.put(id, cached);   
            }
        }
        
        return cached;
    }
    
    /**
     * Load string values from a file using the given connection. The connection
     * isn't closed by this method.
     * 
     * @param id identifier for the group of string values.
     * @return the string values or <code>null</code> if none were available.
     */
    private static String[] loadGroupValues(String id, FileConnection conn) {
        DataInputStream dis = null;
        
        try {

            if (conn.exists()) {
                dis = conn.openDataInputStream();
                return readArray(dis);
            } 
            
        } catch (IOException ioe) {
            if (LogLevel.level >= Log.ERROR) Log.log("Failed to load values file", ioe);
        } finally {
            
            if (dis != null) {
                try {
                    dis.close();
                } catch (IOException e) {
                }
            }
        }
        
        return null;
    }
    
    /**
     * Adds a value to a group of string values identified by parameter
     * <code>id</code>.
     * 
     * @param id the group identifier.
     * @param value the new value to add to the group.
     * @param ignoreCase if <code>true</code> then when searching for an existing 
     *                   value the case will be ignored.
     */
    public synchronized static String[] addPreviousValue(String id, String value, boolean ignoreCase) {

        if (value == null) {
            throw new NullPointerException();
        }
        
        String[] all = (String[]) cachedGroups.get(id);
        FileConnection conn = null;
        boolean openedOnce = false;
        
        if (all == null) {
            // Nothing cached, check if values exist in a file.
            conn = openFile(id, Connector.READ_WRITE);
            openedOnce = true;
            
            if (conn != null) {
                all = loadGroupValues(id, conn);   
            }
        }
        
        String[] newAll = null;
        boolean save = true;
        
        if (value.length() > 0 && !valueExists(all, value, ignoreCase)) {
            newAll = appendValue(all, value);
        } else {
            save = false;
            newAll = all;
        }         

        try {
            
            if (save && newAll != null) {

                if (!openedOnce) {
                    // Previous values are from cache, opening connection hasn't been tried yet.
                    conn = openFile(id, Connector.READ_WRITE);
                }
                
                if (conn != null) {
                    saveGroupValues(newAll, conn);   
                } 
            }
            
        } finally {
            try {
                if (conn != null) conn.close();
            } catch (IOException e) {
            }
        }
        
        if (newAll != null) {
            cachedGroups.put(id, newAll);    
        }

        return newAll;
    }
    
    public static final boolean valueExists(String[] set, String value, boolean ignoreCase) {
        
        if (set != null) {
            
            for (int i = 0; i < set.length; i++) {
                
                if ((ignoreCase && value.equalsIgnoreCase(set[i])) || 
                    (!ignoreCase && value.equals(set[i]))) 
                {
                    return true;
                }
            }
        }
        
        return false;
    }
    
    private static final String[] readArray(DataInputStream dis) throws IOException {
        int amount = dis.readInt();
        String[] ret = new String[amount];
        
        while (amount > 0) {
            ret[ret.length - amount] = dis.readUTF();
            amount--;
        }
        
        return ret;
    }
    
    /**
     * Saves the given string values to a file using the specified {@link FileConnection}.
     * The connection isn't closed by this method.<p>
     * 
     * If parameter <code>values</code> value is <code>null</code> then the
     * file is only deleted if it exists.<p>
     */
    private static final void saveGroupValues(String[] values, FileConnection conn) {
        DataOutputStream dos = null;
        
        try {

            // Always overwrite an existing file.
            if (conn.exists()) {
                conn.delete();
            }

            if (values != null) {
                conn.create();
                dos = conn.openDataOutputStream();
                writeArray(values, dos);    
            }
            
        } catch (IOException ioe) {
            if (LogLevel.level >= Log.ERROR) Log.log("Failed to save values to file", ioe);
        } finally {
            
            if (dos != null) {
                try {
                    dos.close();
                } catch (IOException e) {
                }
            }
        }
    }
    
    private static final void writeArray(String[] values, DataOutputStream dos) throws IOException {
        dos.writeInt(values.length);
        
        for (int i = 0; i < values.length; i++) {
            dos.writeUTF(values[i]);
        }
        
        dos.flush();
    }
    
    private static final FileConnection openFile(String id, int type) {
        
        for (int i = 0; i < LOCATION_PROPS.length; i++) {
            String property = LOCATION_PROPS[i];
            
            try {
                String dir = System.getProperty(property);
                FileConnection conn = (FileConnection) Connector.open(dir, Connector.READ);
                
                if (conn.exists()) {
                    conn.close();
                    String file = dir + getFilename(id); 
                    conn = (FileConnection) Connector.open(file, type);
                } else {
                    // The directory doesn't exist. Try another one.
                    continue;
                }
                
                return conn;
            } catch (SecurityException se) {
                // Stop immediately because the user probably didn't allow access.
                if (LogLevel.level >= Log.ERROR) Log.log("Failed to open values file", se);
                return null;
            } catch (IOException ioe) {
                if (LogLevel.level >= Log.ERROR) Log.log("Failed to open values file", ioe);
                // Try if the other locations would result in success.
            }
        }

        // File could not be opened.
        return null;
    }
    
    private static final String getFilename(String id) {
        return id + EXTENSION;
    }
    
    public static final String[] appendValue(String[] values, String value) {
        
        if (values == null) {
            return new String[] {value};
        } else {
            String[] ret = new String[values.length + 1];
            System.arraycopy(values, 0, ret, 0, values.length);
            ret[ret.length - 1] = value;
            return ret;
        }
    }

}
