package com.nokia.nfc.app.util;

import java.util.Calendar;

/**
 * Utility methods used by JSR-257 implementation.
 */
public class Tools {

    /** Hexadecimal string prefix. */
    private static final String HEX_PREFIX = "0x";
    
    /**
     * Returns the entire array as a hexadecimal string that starts with 
     * the prefix '0x'.
     * 
     * @param data bytes to be converted into a hex string
     * @return a hexadecimal string.
     */
    public static String getAsHexString(byte[] data) {
        return getAsHexString(data, true);
    }
    
    /**
     * Returns the entire array as a hexadecimal string.
     * 
     * @param data bytes to be converted into a hex string
     * @param prefix if true then string will start with '0x'
     * @return a hexadecimal string.
     */
    public static final String getAsHexString(byte[] data, boolean prefix) {
        final int len = (data == null ? 0 : data.length);
        return getAsHexString(data, 0, len, prefix);
    }

    /**
     * Returns the given integer as a hexadecimal string that starts with 
     * the prefix '0x'.
     * 
     * @param value the integer value.
     * @return a hexadecimal string.
     */
    public static final String getAsHexString(int value) {
        return getAsHexString(value, true);
    }
    
    /**
     * Returns the given integer as a hexadecimal string.
     * 
     * @param value the integer value.
     * @param prefix prefix if true then string will start with '0x'
     * @return a hexadecimal string.
     */
    public static final String getAsHexString(int value, boolean prefix) {
        return (prefix ? HEX_PREFIX : "") + Integer.toHexString(value);
    }
    
    /**
     * Checks if the given value falls into the unsigned short value range
     * (0-65535). This range is used when defining block numbers and amounts
     * in the read and write requests.
     */
    public static final boolean isShort(int value) {
        return (value >= 0 && value <= 65535);
    }

    /**
     * Returns the given byte value as a hexadecimal string.
     * 
     * @param b the byte value
     * @param prefix prefix if true then string will start with '0x'
     * @return a hexadecimal string.
     */
    public static final String getAsHexString(byte b, boolean prefix) {
        StringBuffer sb = new StringBuffer();
        
        if (prefix) {
            sb.append(HEX_PREFIX);
        }
        
        appendByte(b, sb);
        return sb.toString();
    }
    
    /**
     * Returns the given byte array as a hexadecimal string without
     * splitting it to multiple lines. The returned string will start with
     * the prefix '0x'. 
     * 
     * @param data bytes to be converted into a hex string
     * @param offset offset into the data array
     * @param len maximum length to be read starting from offset (will stop at array end if too big)
     * @return a hexadecimal string.
     */
    public static final String getAsHexString(byte[] data, int offset, int len) {
        return getAsHexString(data, offset, len, true);
    }
    
    /**
     * Returns the given byte array as a hexadecimal string without
     * splitting it to multiple lines. 
     * 
     * @param data bytes to be converted into a hex string
     * @param offset offset into the data array
     * @param len maximum length to be read starting from offset (will stop at array end if too big)
     * @param prefix if true then string will start with '0x'
     * @return a hexadecimal string.
     */
    public static final String getAsHexString(byte[] data, int offset, int len, boolean prefix) {
        StringBuffer sb = new StringBuffer();

        if (prefix) {
            sb.append(HEX_PREFIX);   
        }

        if (data != null) {
            
            if (offset + len > data.length) {
                len = data.length - offset;
            }
            
            for (int i = offset; i < offset + len; i++) {
                byte b = data[i];
                appendByte(b, sb);
            }
        }
        
        return sb.toString();
    }

    /**
     * Returns the given byte array as a hexadecimal string that is split into
     * lines of 16 bytes. Each line starts with a offset number. 
     */
    public static final String getAsHexStringL(byte[] data) {
        return getAsHexStringL(data, 16, true);
    }
    
    /**
     * Returns the given byte array as a hexadecimal string that is split into
     * lines of arbitrary number of bytes. 
     * 
     * @param perLine how many bytes there should be per line.
     * @param showAddr if true then displays byte index in the beginning of lines. 
     * @return given bytes as a hexadecimal string.
     */
    public static final String getAsHexStringL(byte[] data, int perLine, boolean showAddr) {
        StringBuffer sb = new StringBuffer();
        int count = 0;
        appendAddress(count, sb);
        
        for (int i = 0; i < data.length; i++) {
            appendByte(data[i], sb);
            
            if (i < (data.length - 1)) {
                count++;
                
                if ((count % perLine) == 0) {
                    sb.append('\n');
                    
                    if (showAddr) {
                        appendAddress(count, sb);    
                    }
                    
                } else {
                    sb.append(' ');
                }
   
            }
            
        }
        
        return sb.toString();
    }
    
    /**
     * Constructs a byte array from the given hexadecimal string.
     * The string may begin with the prefix '0x'.
     * 
     * @param hexString the hexadecimal string.
     * @return a byte array. Never <code>null</code>.
     * 
     * @throws NumberFormatException if the string has invalid characters.
     */
    public static final byte[] readHexString(String hexString) {
        
        if (hexString == null || hexString.length() == 0 || hexString.equals(HEX_PREFIX)) {
            return new byte[] {};
        } else {
            
            if (hexString.startsWith(HEX_PREFIX)) {
                hexString = hexString.substring(2);
            }
            
            byte[] data = new byte[hexString.length() / 2];
            
            for (int i = 0; i < data.length; i++) {
                data[i] = (byte) (Integer.parseInt(hexString.substring(i * 2, i * 2 + 2), 16) & 0xFF);
            }
            
            return data;
        }
    }
    
    /**
     * Appends a address (array offset) to the given string. 
     */
    private static final void appendAddress(int addr, StringBuffer sb) {
        appendByte(addr, sb);
        sb.append("-: ");
    }
    
    /**
     * Appends a byte to the given string.
     */
    private static final void appendByte(int byteValue, StringBuffer sb) {
        String bStr = Integer.toHexString(byteValue & 0xFF);
        
        if (bStr.length() < 2) {
            sb.append('0');
        } 
        
        sb.append(bStr);
    }
    
    /**
     * Reads an (signed) integer from the given byte array starting at the specified
     * array index.
     * 
     * @param data byte array containing the 4 integer bytes.
     * @param offset beginning of the integer value in the array.
     * @param littleEndian true if little endian byte order is used. 
     *                     If false then the integer is read using big endian.
     * @return the read integer value.
     */
    public static final int readInt(byte[] data, int offset, boolean littleEndian) {
        
        if (littleEndian) {
            return (((data[offset + 3] & 0xff) << 24) | 
                    ((data[offset + 2] & 0xff) << 16) |
                    ((data[offset + 1] & 0xff) << 8) | 
                     (data[offset + 0] & 0xff));
        } else {
            return (((data[offset + 0] & 0xff) << 24) | 
                    ((data[offset + 1] & 0xff) << 16) |
                    ((data[offset + 2] & 0xff) << 8) | 
                     (data[offset + 3] & 0xff));
        }
    }
    
    /**
     * Reads an integer from the given byte array starting at the specified
     * array index using big endian byte order.
     * 
     * @return the read integer value.
     */
    public static final int readInt(byte[] data, int offset) {
        return readInt(data, offset, false);
    }

    /**
     * Reads a (signed) short integer from the byte array.
     * 
     * @param data byte array containing the 2 integer bytes.
     * @param offset beginning of the short integer value in the array.
     * @param littleEndian true if little endian byte order is used. 
     *                     If false then the integer is read using big endian.
     * @return the read short integer value.
     */
    public static final int readShort(byte[] data, int offset, boolean littleEndian) {
        
        if (littleEndian) {
            return ((data[offset + 1] << 8) | (data[offset] & 0xFF)) & 0xFFFF;
        } else {
            return ((data[offset] << 8) | (data[offset + 1] & 0xFF)) & 0xFFFF;
        }
    }
    
    /**
     * Reads a (signed) short integer from the byte array using 
     * big endian byte order.
     */
    public static final int readShort(byte[] data, int offset) {
        return readShort(data, offset, false);
    }
    
    /**
     * Writes a short integer (2B) to the byte array.
     * 
     * @param data a byte array.
     * @param offset defines the offset of the short integer in the array.
     * @param value the value to be written.
     * @param littleEndian if true then little endianness is used. Otherwise big endian.
     */
    public static final void writeShort(byte[] data, int offset, int value, boolean littleEndian) {
        
        if (littleEndian) {
            data[offset + 1] = (byte)((value >> 8) & 0xFF);
            data[offset + 0] = (byte)(value & 0xFF);   
        } else {
            data[offset + 0] = (byte)((value >> 8) & 0xFF);
            data[offset + 1] = (byte)(value & 0xFF);
        }
    }
    
    /**
     * Writes a short integer (2B) to the byte array using big endian
     * byte order.
     */
    public static final void writeShort(byte[] data, int offset, int value) {
        writeShort(data, offset, value, false);
    }
    
    /**
     * Writes a integer (4B) to the byte array.
     * 
     * @param data a byte array.
     * @param offset defines the offset of the integer in the array.
     * @param value the value to be written.
     * @param littleEndian if true then little endianness is used. Otherwise big endian.
     */
    public static final void writeInt(byte[] data, int offset, long value, boolean littleEndian) {
        
        if (littleEndian) {
            data[offset + 3] = (byte)((value >> 24) & 0xFF);
            data[offset + 2] = (byte)((value >> 16) & 0xFF);
            data[offset + 1] = (byte)((value >> 8) & 0xFF);
            data[offset + 0] = (byte)(value & 0xFF);
        } else {
            data[offset + 0] = (byte)((value >> 24) & 0xFF);
            data[offset + 1] = (byte)((value >> 16) & 0xFF);
            data[offset + 2] = (byte)((value >> 8) & 0xFF);
            data[offset + 3] = (byte)(value & 0xFF);
        }
    }
    
    /**
     * Writes a integer (4B) to the byte array using big endian
     * byte order.
     */
    public static final void writeInt(byte[] data, int offset, long value) {
        writeInt(data, offset, value, false);
    }
    
    /**
     * Appends a timestamp string into the given string buffer.
     * 
     * @param timestamp 'HHmm.ss ' or 'yyyyMMdd_HHmm.ss ' if date is true.
     * @param date if true then day of month, month and year are to be set.
     */
    public static final void setTimeStamp(StringBuffer timestamp, boolean date) {
        // '(ddMMyyyy_)HHmm.ss '
        Calendar calendar = Calendar.getInstance(); 
        int hour = calendar.get(Calendar.HOUR_OF_DAY);
        int min = calendar.get(Calendar.MINUTE);
        int sec = calendar.get(Calendar.SECOND);
        int year = calendar.get(Calendar.YEAR);
        int month = calendar.get(Calendar.MONTH) + 1;
        int dom = calendar.get(Calendar.DAY_OF_MONTH);
        int offset = 0;
        
        if (date) {
            appendNumber(year, 4, timestamp, 0);
            appendNumber(month, 2, timestamp, 4);
            appendNumber(dom, 2, timestamp, 6);
            timestamp.setCharAt(8, '_');
            offset += 9;
        }
        

        appendNumber(hour, 2, timestamp, offset);
        appendNumber(min, 2, timestamp, offset + 2);
        appendNumber(sec, 2, timestamp, offset + 5);
        timestamp.setCharAt(offset + 4, '.');
    }
    
    /**
     * Appends the given amount of digits to the string buffer. Digits are
     * read from the given number parameter value.
     */
    private static final void appendNumber(int number, int digits, StringBuffer sb, int offset) {
        while (digits > 0) {
            sb.setCharAt(offset + digits - 1, (char)(48 + (number % 10)));
            number /= 10;
            digits--;
        }
    }
    
    /**
     * Returns a copy of the given byte array.
     */
    public static final byte[] arrayCopy(byte[] src) {
        
        if (src == null) {
            return null;
        } else {
            byte[] ret = new byte[src.length];
            System.arraycopy(src, 0, ret, 0, src.length);
            return ret;
        }
    }
    
    /**
     * Returns true if the two given byte arrays have matching content
     * or if they both are <code>null</code>.
     * 
     * @param aOffset defines where to start comparing bytes in the <code>a</code> array.
     * @param bOffset defines where to start comparing bytes in the <code>b</code> array.
     * @param len defines how many bytes should be compared.
     */
    public static final boolean equals(byte[] a, byte[] b, int aOffset, int bOffset, int len) {
        
        if (a == null && b == null) {
            return true;
        } else if (a != null && b != null && (aOffset + len) <= a.length && (bOffset + len) <= b.length) {
            
            for (int i = 0; i < len; i++) {
                if (a[aOffset + i] != b[bOffset + i]) {
                    return false;
                }
            }
            
            return true;
        } else {
            return false;   
        }
    }
    
    /**
     * Returns true if the two given byte arrays are equal in length and
     * in their contents match, or if they both are <code>null</code>.
     */
    public static final boolean equals(byte[] a, byte[] b) {
        
        if (a == null && b == null) {
            return true;
        } else if (a != null && b != null && a.length == b.length) {
            
            for (int i = 0; i < a.length; i++) {
                if (a[i] != b[i]) {
                    return false;
                }
            }
            
            return true;
        } else {
            return false;   
        }
    }
    
    /**
     * Compares two sets of bytes. Both sets are in the given byte array
     * at specified indices. No bounds checking is done for the parameters.
     * 
     * @param bytes array that contains the bytes to compare.
     * @param aOffset offset of byte set A.
     * @param bOffset offset of byte set B.
     * @param len amount of bytes to compare.
     * @param inverted set to <code>true</code> if set B bytes should invert equal set A.
     * 
     * @return <code>true</code> if the two sets are equal.
     */
    public static final boolean byteEquals(byte[] bytes, int aOffset, int bOffset, int len, boolean inverted) {
        
        for (int i = 0; i < len; i++) {
            if ((inverted && bytes[aOffset + i] != (~bytes[bOffset + i])) || 
                (!inverted && (bytes[aOffset + i] != bytes[bOffset + i]))) 
            {
                return false;
            }
        }
        
        return true;
    }

}
